/*******************************************************************************
 * Copyright (c) 2009 Information-technology Promotion Agency, Japan.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package benten.cat.ui.dialogs;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.StatusDialog;
import org.eclipse.jface.text.TextViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

import benten.cat.ui.viewers.TranslationSourceViewer;
import benten.ui.UiStatus;
import benten.ui.views.UiToolkit;

/**
 * ノートを編集するためのダイアログです。
 */
public class NoteDialog extends StatusDialog implements ModifyListener {

	/** ノート・エントリー・クラス */
	public static class NoteDialogEntry {

		/** 担当 */
		public String from;
		/** 内容 */
		public String content;

		/**
	 * コンストラクター。
		 * @param from 担当
		 * @param content 内容
		 */
		public NoteDialogEntry(String from, String content) {
			this.from = from;
			this.content = content;
		}

		/*
		 * (非 Javadoc)
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		@Override
		public boolean equals(Object obj) {
			if (obj != null || obj instanceof NoteDialogEntry) {
				final NoteDialogEntry note = (NoteDialogEntry) obj;
				return note.from.equals(from) && note.content.equals(content);
			}
			return false;
		}
	}

	private final UiToolkit toolkit = new UiToolkit();
	private final NoteDialogEntry noteEntry;
	private Text fromText;
	private TextViewer textTextViewer;

	/**
	 * コンストラクター。
	 * @param parent 親シェル
	 * @param title タイトル
	 * @param noteEntry 編集対象となるノート・エントリー
	 */
	public NoteDialog(Shell parent, String title, NoteDialogEntry noteEntry) {

		super(parent);
		this.noteEntry = noteEntry;
		setTitle(title);
		setHelpAvailable(false);
	}

	/*
	 * (非 Javadoc)
	 * @see org.eclipse.jface.dialogs.Dialog#createDialogArea(org.eclipse.swt.widgets.Composite)
	 */
	@Override
	protected Control createDialogArea(Composite parent) {

		final Composite composite = (Composite) super.createDialogArea(parent);
		final Composite c = toolkit.createComposite(composite, 2);
		c.setLayoutData(new GridData(GridData.FILL_BOTH));

		toolkit.createLabel(c, "担当(&C):");
		fromText = new Text(c, SWT.SINGLE | SWT.BORDER);
		fromText.setText(noteEntry.from);
		fromText.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		fromText.addModifyListener(this);

		toolkit.createLabel(c, "内容(&T):");
		final GridData gd = new GridData(GridData.FILL_BOTH);
		gd.heightHint = 60;
		gd.widthHint = 400;
		textTextViewer = new TranslationSourceViewer(c, true);
		textTextViewer.getControl().setLayoutData(gd);
		textTextViewer.getDocument().set(noteEntry.content);
		textTextViewer.getTextWidget().addModifyListener(this);

		return composite;
	}

	/*
	 * (非 Javadoc)
	 * @see org.eclipse.jface.dialogs.Dialog#okPressed()
	 */
	@Override
	protected void okPressed() {

		final NoteDialogEntry editedNoteEntry = getNoteDialogEntry();
		if (editedNoteEntry.equals(noteEntry)) {
			cancelPressed(); // 変更がない場合はキャンセル扱いにする
		} else {
			noteEntry.from = editedNoteEntry.from;
			noteEntry.content = editedNoteEntry.content;
			super.okPressed();
		}
	}

	/*
	 * (非 Javadoc)
	 * @see org.eclipse.jface.dialogs.Dialog#isResizable()
	 */
	@Override
	protected boolean isResizable() {
		return true;
	}

	/*
	 * (非 Javadoc)
	 * @see org.eclipse.swt.events.ModifyListener#modifyText(org.eclipse.swt.events.ModifyEvent)
	 */
	public void modifyText(ModifyEvent e) {

		Status status = UiStatus.OK_STATUS;
		final NoteDialogEntry editedNoteEntry = getNoteDialogEntry();
		if (editedNoteEntry.from.equals("") && editedNoteEntry.content.equals("")) {
			status = new UiStatus(IStatus.ERROR, "担当、内容のいずれかを入力してください。");
		}
		updateStatus(status);
	}

	private NoteDialogEntry getNoteDialogEntry() {
		return new NoteDialogEntry(fromText.getText(), textTextViewer.getDocument().get());
	}
}
