/*******************************************************************************
 * Copyright (c) 2009 Information-technology Promotion Agency, Japan.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package benten.cat.ui.viewers;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jface.layout.TableColumnLayout;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.text.DocumentEvent;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IDocumentListener;
import org.eclipse.jface.text.ITextListener;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.text.ITextViewer;
import org.eclipse.jface.text.TextEvent;
import org.eclipse.jface.text.source.SourceViewer;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ColumnWeightData;
import org.eclipse.jface.viewers.ContentViewer;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IContentProvider;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.IPostSelectionProvider;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.ProgressBar;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.forms.events.ExpansionAdapter;
import org.eclipse.ui.forms.events.ExpansionEvent;
import org.eclipse.ui.views.contentoutline.IContentOutlinePage;
import org.eclipse.wst.sse.core.StructuredModelManager;
import org.eclipse.wst.sse.core.internal.provisional.IStructuredModel;
import org.eclipse.wst.sse.core.internal.provisional.text.IStructuredDocumentRegion;
import org.eclipse.wst.sse.core.internal.provisional.text.ITextRegion;
import org.eclipse.wst.sse.ui.internal.contentoutline.ConfigurableContentOutlinePage;
import org.eclipse.wst.xml.core.internal.provisional.document.IDOMDocument;
import org.eclipse.wst.xml.core.internal.provisional.document.IDOMModel;
import org.eclipse.wst.xml.core.internal.provisional.document.IDOMNode;
import org.eclipse.wst.xml.core.internal.regions.DOMRegionContext;
import org.eclipse.wst.xml.ui.internal.tabletree.IDesignViewer;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import benten.cat.glossary.ui.editor.GlossaryConfiguration;
import benten.cat.ui.CatUiPlugin;
import benten.cat.ui.dialogs.NoteDialog;
import benten.cat.ui.dialogs.NoteDialog.NoteDialogEntry;
import benten.core.dom.AltTransDelegate;
import benten.core.dom.ElementDelegate;
import benten.core.dom.NoteDelegate;
import benten.core.dom.TransUnitDelegate;
import benten.core.model.BentenXliff;
import benten.core.model.HelpTransUnitId;
import benten.ui.views.AsyncSelectionChangeAdapter;
import benten.ui.views.TableLabelProvider;
import benten.ui.views.UiToolkit;

/**
 * XLIFF エディターの翻訳単位タブに表示するビューアーです。
 * 
 * ★基本設計「翻訳支援機能: 翻訳中間形式エディター機能: 翻訳単位編集」に対応します。
 * ★基本設計「翻訳支援機能: 翻訳中間形式エディター機能: 翻訳単位選択」に対応します。
 * TODO 選択、編集が、両方ともここに該当しますでしょうか?
 */
@SuppressWarnings( { "restriction" })
public class TransUnitViewer extends ContentViewer implements IDesignViewer {

	public static enum StoreKey {
		DEFAULT_NOTE_FROM, EXPAND_FILTER_STATE_CHECKS, EXPAND_ALT_TRANS_TABLE, EXPAND_NOTE_LIST_TABLE
	}

	private final UiToolkit toolkit = new UiToolkit();
	private final IEditorPart editorPart;
	private final SashForm sashForm;

	private TableViewer transUnitTableViewer;
	private Label transUnitCountLabel;
	private Button joinTransUnitButton;
	private Button prevTransUnitButton;
	private Button nextTransUnitButton;

	private Text idText;
	private Combo stateCombo;
	private Button notTranslateCheck;
	private Button updateTmCheck;

	private SourceViewer sourceTextViewer;
	private SourceViewer targetTextViewer;
	private SourceViewer altSourceTextViewer;
	private SourceViewer altTargetTextViewer;

	private Button copyFromSourceButton;
	private Button copyFromAltTransButton;

	private TableViewer altTransTableViewer;
	private Text originLabel;
	private ProgressBar altTransQualityBar;
	private Label altTransCountLabel;
	private Button prevAltTransButton;
	private Button nextAltTransButton;

	private TableViewer noteTableViewer;
	private Button newNoteButton;
	private Button editNoteButton;
	private Button removeNoteButton;

	private TransUnitDelegate selectedTransUnit;
	private NoteDelegate selectedNote;

	private final Set<String> checkedStateItems = new HashSet<String>();
	private boolean internalSelectionChange;

	/**
	 * コンストラクター。
	 * @param parent 親コンポジット
	 * @param editorPart エディター・パーツ
	 */
	public TransUnitViewer(Composite parent, IEditorPart editorPart) {

		this.editorPart = editorPart;

		sashForm = new SashForm(parent, SWT.BORDER | SWT.VERTICAL | SWT.V_SCROLL | SWT.H_SCROLL);
		createTransUnitListGroup(sashForm);
		createTransUnitGroup(sashForm);
		sashForm.setWeights(new int[] { 35, 65 });

		setContentProvider(new IContentProvider() {
			public void dispose() {
			}

			public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
			}
		});
	}

	private void createTransUnitListGroup(Composite parent) {

		final Composite c = toolkit.createComposite(parent, 1, 10);

		final Group g = new Group(c, SWT.BOLD);
		g.setLayout(toolkit.createGridLayout(1, 5));
		g.setLayoutData(new GridData(GridData.FILL_BOTH));
		g.setText("翻訳単位リスト");

		createStateItemGroup(g);
		createTransUnitTableViewer(g);
		createTransUnitButtons(g);
	}

	private void createStateItemGroup(final Composite parent) {

		final PersistenceExpandableComposite expandable = new PersistenceExpandableComposite(parent, "フィルターする状態(&D):");
		final Composite outer = new Composite(expandable, SWT.BORDER);
		outer.setLayout(new FillLayout());

		final Composite c = toolkit.createComposite(outer, 6, 5);

		for (final String stateItem : BentenXliff.getStateItems()) {

			checkedStateItems.add(stateItem);

			final Button checkButton = new Button(c, SWT.CHECK);
			checkButton.setData(stateItem);
			checkButton.setText(stateItem.equals("") ? "なし" : stateItem);
			checkButton.setSelection(true);

			checkButton.addSelectionListener(new SelectionAdapter() {
				@Override
				public void widgetSelected(SelectionEvent e) {
					final Button b = (Button) e.widget;
					final String selectedStateItem = (String) b.getData();
					if (b.getSelection()) {
						checkedStateItems.add(selectedStateItem);
					} else {
						checkedStateItems.remove(selectedStateItem);
					}
					refresh();
				}
			});
		}

		expandable.loadExpandedSetting(StoreKey.EXPAND_FILTER_STATE_CHECKS.name());
		expandable.setClient(outer);
		expandable.addExpansionListener(new ExpansionAdapter() {
			@Override
			public void expansionStateChanged(ExpansionEvent e) {
				packTable(parent, e.getState(), transUnitTableViewer);
			}
		});
	}
	private void createTransUnitTableViewer(Composite parent) {

		final Composite tableParent = new Composite(parent, SWT.NONE);
		tableParent.setLayoutData(new GridData(GridData.FILL_BOTH));

		final Table table = new Table(tableParent, SWT.BORDER | SWT.FULL_SELECTION);
		table.setHeaderVisible(true);
		table.setLinesVisible(true);

		final TableColumnLayout tableColumnLayout = new TableColumnLayout();
		tableParent.setLayout(tableColumnLayout);

		TableColumn tableColumn = new TableColumn(table, SWT.NONE);
		tableColumn.setText("状態");
		tableColumnLayout.setColumnData(tableColumn, new ColumnWeightData(16));

		tableColumn = new TableColumn(table, SWT.NONE);
		tableColumn.setText("ソース");
		tableColumnLayout.setColumnData(tableColumn, new ColumnWeightData(52));

		tableColumn = new TableColumn(table, SWT.NONE);
		tableColumn.setText("ターゲット");
		tableColumnLayout.setColumnData(tableColumn, new ColumnWeightData(52));

		transUnitTableViewer = new TableViewer(table);
		transUnitTableViewer.setContentProvider(new ArrayContentProvider());
		transUnitTableViewer.setLabelProvider(new TableLabelProvider() {
			public String getColumnText(Object element, int columnIndex) {

				final TransUnitDelegate transUnit = new TransUnitDelegate((IDOMDocument) getInput(), (Element) element);
				switch (columnIndex) {
				case 0:
					return transUnit.getTargetState();
				case 1:
					return transUnit.getSource();
				case 2:
					return transUnit.getTarget();
				}
				return null;
			}
		});
		transUnitTableViewer.addPostSelectionChangedListener(new AsyncSelectionChangeAdapter() {
			@Override
			public void selectionChanged(IStructuredSelection selection) {
				if (selection.isEmpty()) {
					return;
				}
				selectedTransUnit = new TransUnitDelegate((IDOMDocument) getInput(), (Element) selection
						.getFirstElement());
				final List<NoteDelegate> noteList = selectedTransUnit.getNoteList();
				final List<AltTransDelegate> altTransList = selectedTransUnit.getAltTransList();

				noteTableViewer.setInput(ElementDelegate.toElementList(noteList));
				idText.setText(selectedTransUnit.getId());
				stateCombo.setText(selectedTransUnit.getTargetState());
				notTranslateCheck.setSelection(!selectedTransUnit.isTranslate());
				updateTmCheck.setSelection(!selectedTransUnit.isContextGourpTmOmit());
				sourceTextViewer.getDocument().set(selectedTransUnit.getSource());
				targetTextViewer.getDocument().set(selectedTransUnit.getTarget());

				setEnabledTransUnit(true);
				updateTranslateCheck(notTranslateCheck.getSelection());

				if (altTransList.size() == 0) {
					clearSelectedAltTrans();
				} else {
					altTransTableViewer.setInput(ElementDelegate.toElementList(altTransList));
					altTransTableViewer.getTable().setSelection(0);
					setSelectedAltTrans(altTransList.get(0));
				}
				updateTransUnitButtons();
				updateAltTransButtons();
				updateNoteButtons();
				updateOutlineView(selection);
			}
		});
	}

	private void createTransUnitGroup(Composite parent) {

		final Composite c = new Composite(parent, SWT.NONE);
		c.setLayout(toolkit.createGridLayout(1, 10));

		final Group g = new Group(c, SWT.NONE);
		g.setLayout(toolkit.createGridLayout(1, 5));
		g.setLayoutData(new GridData(GridData.FILL_BOTH));
		g.setText("翻訳単位");

		createTransUnitAttributes(g);
		createTransUnitText(g);
		createAltTransButtons(g);
		createAltTransTableViewer(g);
		createNoteTableViewer(g);
	}

	private void createTransUnitAttributes(Composite parent) {

		Composite c = toolkit.createComposite(parent, 2);
		c.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));

		toolkit.createLabel(c, "ID:", 50);
		idText = new Text(c, SWT.SINGLE | SWT.BORDER | SWT.READ_ONLY);
		idText.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));

		c = toolkit.createComposite(parent, 4);
		c.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));

		toolkit.createLabel(c, "状態(&S):", 50);
		stateCombo = new Combo(c, SWT.READ_ONLY);
		stateCombo.setItems(BentenXliff.getStateItems());
		stateCombo.addModifyListener(new ModifyListener() {
			public void modifyText(ModifyEvent e) {
				if (selectedTransUnit == null || !targetTextViewer.isEditable()) {
					return;
				}
				final String state = ((Combo) e.getSource()).getText();
				if (selectedTransUnit.setTargetState(state)) {
					transUnitTableViewer.update(selectedTransUnit.getElement(), null);
				}
			}
		});

		GridData gd = new GridData();
		gd.horizontalAlignment = SWT.RIGHT;
		gd.grabExcessHorizontalSpace = true;
		notTranslateCheck = new Button(c, SWT.CHECK);
		notTranslateCheck.setLayoutData(gd);
		notTranslateCheck.setText("翻訳対象外(&I)");
		notTranslateCheck.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				final boolean checked = ((Button) e.widget).getSelection();
				selectedTransUnit.setTranslate(!checked);
				updateTranslateCheck(checked);
			}
		});

		gd = new GridData();
		gd.horizontalAlignment = SWT.RIGHT;
		gd.horizontalIndent = 30;
		updateTmCheck = new Button(c, SWT.CHECK);
		updateTmCheck.setLayoutData(gd);
		updateTmCheck.setText("翻訳メモリーに反映する(&M)");
		updateTmCheck.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				final boolean checked = ((Button) e.widget).getSelection();
				selectedTransUnit.setContextGroupTmOmit(!checked);
			}
		});
	}

	private void createTransUnitText(Composite parent) {
		final Composite c = toolkit.createComposite(parent, 3);
		c.setLayoutData(new GridData(GridData.FILL_BOTH));

		sourceTextViewer = createTextViewer(c, "ソース:");
		new Label(c, SWT.NONE);
		altSourceTextViewer = createTextViewer(c, "選択された代替翻訳のソース:");

		targetTextViewer = createTargetTextViewer(c, "ターゲット(&G):");

		final Composite copyButtonOuter = toolkit.createComposite(c, 1);
		copyButtonOuter.setLayoutData(new GridData(GridData.FILL_VERTICAL));

		final Label label = toolkit.createLabel(copyButtonOuter, "");
		final GridData gd = new GridData();
		gd.heightHint = 16;
		label.setLayoutData(gd);

		altTargetTextViewer = createTextViewer(c, "選択された代替翻訳のターゲット:");
		copyFromAltTransButton = createCopyButton(copyButtonOuter, SWT.LEFT, altTargetTextViewer);
	}

	private Button createCopyButton(Composite parent, int direction, final ITextViewer from) {
		final Button button = new Button(parent, SWT.ARROW | direction);
		button.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				final String copyFrom = from.getDocument().get();
				targetTextViewer.getDocument().set(copyFrom);
			}
		});
		return button;
	}

	private SourceViewer createTextViewer(Composite parent, String text) {
		final Composite c = toolkit.createComposite(parent, 1, 1);
		c.setLayoutData(new GridData(GridData.FILL_BOTH));

		final Label label = toolkit.createLabel(c, text, 200);
		final GridData gd = (GridData) label.getLayoutData();
		gd.heightHint = 16;
		gd.verticalAlignment = SWT.BOTTOM;

		final SourceViewer viewer = new TranslationSourceViewer(c, false);
		return viewer;
	}

	private SourceViewer createTargetTextViewer(Composite parent, String text) {
		final Composite c = toolkit.createComposite(parent, 2, 1);
		c.setLayoutData(new GridData(GridData.FILL_BOTH));

		final Label label = toolkit.createLabel(c, text, 200);
		GridData gd = (GridData) label.getLayoutData();
		gd.heightHint = 16;
		gd.verticalAlignment = SWT.BOTTOM;

		copyFromSourceButton = createCopyButton(c, SWT.DOWN, sourceTextViewer);
		copyFromSourceButton.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_END));

		final SourceViewer viewer = new TranslationSourceViewer(c, true);
		gd = (GridData) viewer.getControl().getLayoutData();
		gd.horizontalSpan = 2;
		viewer.getControl().setLayoutData(gd);

		viewer.configure(new GlossaryConfiguration());
		viewer.addTextListener(new ITextListener() {
			public void textChanged(final TextEvent event) {
				if (selectedTransUnit == null || !viewer.isEditable()) {
					return;
				}
				Display.getCurrent().asyncExec(new Runnable() {
					public void run() {
						final String target = event.getDocumentEvent().getDocument().get();
						if (selectedTransUnit.setTarget(target)) {
							transUnitTableViewer.update(selectedTransUnit.getElement(), null);
						}
					}
				});
			}
		});
		return viewer;
	}

	private void createAltTransTableViewer(final Composite parent) {
		final PersistenceExpandableComposite expandable = new PersistenceExpandableComposite(parent, "代替翻訳選択リスト(&L):");

		final GridData gd = new GridData(GridData.FILL_HORIZONTAL);
		gd.heightHint = 200;
		final Composite tableParent = new Composite(expandable, SWT.NONE);
		tableParent.setLayoutData(gd);

		expandable.loadExpandedSetting(StoreKey.EXPAND_ALT_TRANS_TABLE.name());
		expandable.setClient(tableParent);
		expandable.addExpansionListener(new ExpansionAdapter() {
			@Override
			public void expansionStateChanged(ExpansionEvent e) {
				packTable(parent, !e.getState(), altTransTableViewer);
			}
		});

		final Table table = new Table(tableParent, SWT.BORDER | SWT.FULL_SELECTION);
		table.setHeaderVisible(true);
		table.setLinesVisible(true);

		final TableColumnLayout tableColumnLayout = new TableColumnLayout();
		tableParent.setLayout(tableColumnLayout);

		TableColumn tableColumn = new TableColumn(table, SWT.NONE);
		tableColumn.setText("起源");
		tableColumnLayout.setColumnData(tableColumn, new ColumnWeightData(12));

		tableColumn = new TableColumn(table, SWT.RIGHT);
		tableColumn.setText("品質");
		tableColumnLayout.setColumnData(tableColumn, new ColumnWeightData(6));

		tableColumn = new TableColumn(table, SWT.NONE);
		tableColumn.setText("ソース");
		tableColumnLayout.setColumnData(tableColumn, new ColumnWeightData(41));

		tableColumn = new TableColumn(table, SWT.NONE);
		tableColumn.setText("ターゲット");
		tableColumnLayout.setColumnData(tableColumn, new ColumnWeightData(41));

		altTransTableViewer = new TableViewer(table);
		altTransTableViewer.setContentProvider(new ArrayContentProvider());
		altTransTableViewer.setLabelProvider(new TableLabelProvider() {
			public String getColumnText(Object element, int columnIndex) {
				if (!(element instanceof Element)) {
					return null;
				}
				final AltTransDelegate altTrans = new AltTransDelegate((Element) element);
				switch (columnIndex) {
				case 0:
					return altTrans.getOrigin();
				case 1:
					return altTrans.getMatchQuality();
				case 2:
					return altTrans.getSource();
				case 3:
					return altTrans.getTarget();
				}
				return null;
			}
		});

		altTransTableViewer.addPostSelectionChangedListener(new AsyncSelectionChangeAdapter() {
			@Override
			public void selectionChanged(IStructuredSelection selection) {
				if (selection.isEmpty()) {
					return;
				}
				final AltTransDelegate altTrans = new AltTransDelegate((Element) selection.toArray()[0]);
				setSelectedAltTrans(altTrans);
				updateAltTransButtons();
				updateOutlineView(selection);
			}
		});

		final String[] dummyForTableHeight = { "", "", "", "", "" };
		altTransTableViewer.setInput(Arrays.asList(dummyForTableHeight));
	}

	private void createNoteTableViewer(final Composite parent) {

		final PersistenceExpandableComposite expandable = new PersistenceExpandableComposite(parent, "ノート・リスト(&O):");
		final Composite c = toolkit.createComposite(expandable, 2);
		c.setLayoutData(new GridData(GridData.FILL_BOTH));

		expandable.loadExpandedSetting(StoreKey.EXPAND_NOTE_LIST_TABLE.name());
		expandable.setClient(c);
		expandable.addExpansionListener(new ExpansionAdapter() {
			@Override
			public void expansionStateChanged(ExpansionEvent e) {
				packTable(parent, !e.getState(), noteTableViewer);
			}
		});

		final GridData gd = new GridData(GridData.FILL_BOTH);
		gd.heightHint = 30;
		final Composite tableParent = new Composite(c, SWT.NONE);
		tableParent.setLayoutData(gd);

		final Table table = new Table(tableParent, SWT.BORDER | SWT.FULL_SELECTION);
		table.setHeaderVisible(true);
		table.setLinesVisible(true);

		final TableColumnLayout tableColumnLayout = new TableColumnLayout();
		tableParent.setLayout(tableColumnLayout);

		TableColumn tableColumn = new TableColumn(table, SWT.NONE);
		tableColumn.setText("担当");
		tableColumnLayout.setColumnData(tableColumn, new ColumnWeightData(20));

		tableColumn = new TableColumn(table, SWT.NONE);
		tableColumn.setText("内容");
		tableColumnLayout.setColumnData(tableColumn, new ColumnWeightData(80));

		noteTableViewer = new TableViewer(table);
		noteTableViewer.setContentProvider(new ArrayContentProvider());
		noteTableViewer.setLabelProvider(new TableLabelProvider() {
			public String getColumnText(Object element, int columnIndex) {
				final NoteDelegate note = new NoteDelegate((Element) element);
				switch (columnIndex) {
				case 0:
					return note.getFrom();
				case 1:
					return note.getValue();
				}
				return null;
			}
		});
		noteTableViewer.addPostSelectionChangedListener(new AsyncSelectionChangeAdapter() {
			@Override
			public void selectionChanged(IStructuredSelection selection) {
				if (!selection.isEmpty()) {
					selectedNote = new NoteDelegate((Element) selection.getFirstElement());
				}
				updateNoteButtons();
				updateOutlineView(selection);
			}
		});
		noteTableViewer.addDoubleClickListener(new IDoubleClickListener() {
			public void doubleClick(DoubleClickEvent event) {
				editNoteDialog();
			}
		});

		final Composite buttonOuter = toolkit.createComposite(c, 1);

		newNoteButton = toolkit.createButton(buttonOuter, "新規", 60);
		newNoteButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				final IPreferenceStore store = CatUiPlugin.getDefault().getPreferenceStore();
				final String from = store.getString(StoreKey.DEFAULT_NOTE_FROM.name());

				final NoteDialogEntry note = new NoteDialogEntry(from, "");
				final NoteDialog dialog = new NoteDialog(sashForm.getShell(), "ノートの新規登録", note);

				if (dialog.open() == Window.OK) {
					final Node noteElement = selectedTransUnit.appendNote(note.from, note.content);
					noteTableViewer.add(noteElement);
					store.setValue(StoreKey.DEFAULT_NOTE_FROM.name(), note.from);
				}
			}
		});
		editNoteButton = toolkit.createButton(buttonOuter, "編集", 60);
		editNoteButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				editNoteDialog();
			}
		});
		removeNoteButton = toolkit.createButton(buttonOuter, "除去", 60);
		removeNoteButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				selectedTransUnit.removeNote(selectedNote);
				final Table table = noteTableViewer.getTable();
				final int index = table.getSelectionIndex();

				table.remove(index);
				if (index > 0) {
					noteTableViewer
							.setSelection(new StructuredSelection(noteTableViewer.getElementAt(index - 1)), true);
				} else {
					noteTableViewer.setSelection(new StructuredSelection(), true);
				}
			}
		});
	}
	private void editNoteDialog() {
		final NoteDialogEntry note = new NoteDialogEntry(selectedNote.getFrom(), selectedNote.getValue());
		final NoteDialog dialog = new NoteDialog(sashForm.getShell(), "ノートの編集", note);

		if (dialog.open() == Window.OK) {
			selectedNote.setFrom(note.from);
			selectedNote.setValue(note.content);
			noteTableViewer.update(selectedNote.getElement(), null);
		}
	}

	private void createTransUnitButtons(Composite parent) {
		final Composite c = toolkit.createComposite(parent, 5);
		c.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));

		joinTransUnitButton = toolkit.createButton(c, "次の翻訳単位と結合(&J)", 170);

		toolkit.createHorizontalSpacer(c);

		transUnitCountLabel = toolkit.createLabel(c, "", 120);
		transUnitCountLabel.setAlignment(SWT.RIGHT);

		prevTransUnitButton = toolkit.createButton(c, "前の翻訳単位(&U)", 140);
		nextTransUnitButton = toolkit.createButton(c, "次の翻訳単位(&T)", 140);

		joinTransUnitButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				selectedTransUnit.joinNextTransUnit();
				refresh();
				moveTransUnitSelection(0);
			}
		});
		prevTransUnitButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				moveTransUnitSelection(-1);
			}
		});
		nextTransUnitButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				moveTransUnitSelection(1);
			}
		});
	}

	private void moveTransUnitSelection(int move) {
		final Table table = transUnitTableViewer.getTable();
		final int index = table.getSelectionIndex();
		final Object dest = transUnitTableViewer.getElementAt(index + move);
		transUnitTableViewer.setSelection(new StructuredSelection(dest), true);
	}

	private void createAltTransButtons(Composite parent) {

		final Composite c = toolkit.createComposite(parent, 5);
		final GridData gd0 = new GridData(GridData.FILL_HORIZONTAL);
		c.setLayoutData(gd0);

		originLabel = new Text(c, SWT.SINGLE | SWT.READ_ONLY | SWT.RIGHT);
		originLabel.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));

		final GridData gd = new GridData();
		gd.widthHint = 80;
		altTransQualityBar = new ProgressBar(c, SWT.SMOOTH);
		altTransQualityBar.setLayoutData(gd);
		altTransQualityBar.setMinimum(0);
		altTransQualityBar.setMaximum(100);

		altTransCountLabel = toolkit.createLabel(c, "", 60);
		altTransCountLabel.setAlignment(SWT.RIGHT);

		prevAltTransButton = toolkit.createButton(c, "前の代替翻訳(&V)", 140);
		nextAltTransButton = toolkit.createButton(c, "次の代替翻訳(&X)", 140);

		prevAltTransButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				final Table table = altTransTableViewer.getTable();
				final int index = table.getSelectionIndex();
				if (index > 0) {
					altTransTableViewer.setSelection(new StructuredSelection(altTransTableViewer
							.getElementAt(index - 1)), true);
				}
				updateAltTransButtons();
			}
		});
		nextAltTransButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				final Table table = altTransTableViewer.getTable();
				final int index = table.getSelectionIndex();
				if (index < table.getItemCount() - 1) {
					altTransTableViewer.setSelection(new StructuredSelection(altTransTableViewer
							.getElementAt(index + 1)), true);
				}
				updateAltTransButtons();
			}
		});
	}

	//-------------------------------------------------------------------------
	// 可視制御

	private void packTable(Composite parent, boolean addSashTopWeight, TableViewer tableViewer) {

		sashForm.setVisible(false);
		parent.pack();

		final int[] weights = sashForm.getWeights();
		final int sum = weights[0] + weights[1];
		weights[0] += addSashTopWeight ? 10 : -10;
		weights[1] = sum - weights[0];
		sashForm.setWeights(weights);

		// 第一カラム幅を固定にすることで、リサイズ時に横スクロールバーが表示されるのを抑止
		final TableColumn column = tableViewer.getTable().getColumn(0);
		column.setWidth(column.getWidth());

		sashForm.setVisible(true);
	}

	private void setEnabledTransUnit(boolean enabled) {
		stateCombo.setEnabled(enabled);
		notTranslateCheck.setEnabled(enabled);
		updateTmCheck.setEnabled(enabled);
		targetTextViewer.setEditable(enabled);
	}

	private void updateTransUnitButtons() {
		final boolean hasContinue = HelpTransUnitId.hasContinue(selectedTransUnit.getId());
		joinTransUnitButton.setEnabled(hasContinue);

		final Table table = transUnitTableViewer.getTable();
		final int index = table.getSelectionIndex();
		prevTransUnitButton.setEnabled(index > 0);
		nextTransUnitButton.setEnabled(index < table.getItemCount() - 1);

		final int row = table.getSelectionIndex() + 1;
		final String s = row + " / " + table.getItemCount();
		transUnitCountLabel.setText(s);
	}

	private void clearSelectedAltTrans() {
		altSourceTextViewer.getDocument().set("");
		altTargetTextViewer.getDocument().set("");
		altTransTableViewer.setInput(Collections.EMPTY_LIST);
		originLabel.setText("");
		altTransQualityBar.setVisible(false);
		altTransCountLabel.setText("");
	}

	private void setSelectedAltTrans(AltTransDelegate altTrans) {
		altSourceTextViewer.getDocument().set(altTrans.getSource());
		altTargetTextViewer.getDocument().set(altTrans.getTarget());

		final String quality = altTrans.getMatchQuality();
		try {
			altTransQualityBar.setSelection(Integer.parseInt(quality.replace("%", "")));
			altTransQualityBar.setVisible(true);
		} catch (final RuntimeException e) {
			altTransQualityBar.setSelection(0);
			altTransQualityBar.setVisible(false);
		}

		String text = altTrans.getOrigin();
		if (quality != null && !quality.equals("")) {
			text += ": " + quality;
		}
		originLabel.setText(text);

		final Table table = altTransTableViewer.getTable();
		final int row = table.getSelectionIndex() + 1;
		final String s = row + " / " + table.getItemCount();
		altTransCountLabel.setText(s);
	}

	private void updateAltTransButtons() {
		final Table table = altTransTableViewer.getTable();
		final int index = table.getSelectionIndex();
		prevAltTransButton.setEnabled(index > 0);
		nextAltTransButton.setEnabled(index < table.getItemCount() - 1);
	}

	private void updateNoteButtons() {
		final ISelection selection = noteTableViewer.getSelection();
		editNoteButton.setEnabled(!selection.isEmpty());
		removeNoteButton.setEnabled(!selection.isEmpty());
	}

	private void updateTranslateCheck(boolean selection) {
		stateCombo.setEnabled(!selection);
		targetTextViewer.setEditable(!selection);
		copyFromSourceButton.setEnabled(!selection);
		copyFromAltTransButton.setEnabled(!selection);
	}

	private void updateOutlineView(ISelection selection) {
		if (internalSelectionChange) {
			return;
		}
		final ConfigurableContentOutlinePage outline = (ConfigurableContentOutlinePage) editorPart
				.getAdapter(IContentOutlinePage.class);
		if (outline.getConfiguration().isLinkedWithEditor(null)) {
			outline.setSelection(selection);
		}
	}

	//-------------------------------------------------------------------------
	// org.eclipse.wst.xml.ui.internal.tabletree.IDesignViewer の実装

	/**
	 * CTabItem 上に表示するコントロールを取得します。
	 * @return コントロール
	 */
	@Override
	public Control getControl() {
		return sashForm;
	}

	/**
	 * 選択プロバイダーを取得します。
	 * @return 選択プロバイダー
	 */
	public ISelectionProvider getSelectionProvider() {
		return selectionProvider;
	}

	/**
	 * ビューアーのタブ・タイトルを取得します。
	 * @return タブ・タイトル
	 */
	public String getTitle() {
		return "翻訳単位";
	}

	/**
	 * ドキュメントを設定します。
	 * @param document ドキュメント
	 */
	public void setDocument(IDocument document) {

		IStructuredModel model = null;
		try {
			model = StructuredModelManager.getModelManager().getExistingModelForRead(document);
			if (model != null) {
				final IDOMDocument domDoc = ((IDOMModel) model).getDocument();

				// ファイルを開いたときにリフレッシュが必要な状態
				if (domDoc.getFirstChild() == null) {
					final IFile file = ((IFileEditorInput) editorPart.getEditorInput()).getFile();
					file.refreshLocal(0, null);
					model = StructuredModelManager.getModelManager().getModelForRead(file);
				}

				setInput(domDoc);
				refresh();

				// UNDO は XML エディター側の機能を利用するため、ドキュメントの変更が発生した場合は
				// 翻訳単位タブに反映するようにリスナーを設定
				document.addDocumentListener(new IDocumentListener() {
					public void documentAboutToBeChanged(DocumentEvent event) {
					}
					public void documentChanged(DocumentEvent event) {
						stateCombo.setText(selectedTransUnit.getTargetState());
						notTranslateCheck.setSelection(!selectedTransUnit.isTranslate());
						updateTmCheck.setSelection(!selectedTransUnit.isContextGourpTmOmit());
						// 日本語が入力できなくなるため他の方法を検討する必要あり
						//targetTextViewer.getDocument().set(selectedTransUnit.getTarget());
					}
				});
			}
		} catch (final IOException e) {
			CatUiPlugin.getDefault().log(e);

		} catch (final CoreException e) {
			CatUiPlugin.getDefault().log(e);

		} finally {
			if (model != null) {
				model.releaseFromRead();
			}
		}
	}

	//-------------------------------------------------------------------------
	// org.eclipse.jface.viewers.Viewer オーバーライド

	/**
	 * このビューアーをリフレッシュします。
	 * LabelProvider や ContentProvider が設定されたときなどに呼び出されます。
	 */
	@Override
	public void refresh() {

		final IDOMDocument domDoc = (IDOMDocument) getInput();
		if (domDoc == null) {
			return;
		}

		final List<TransUnitDelegate> allTransUnitList = TransUnitDelegate.listOf(domDoc);
		final List<TransUnitDelegate> transUnitList = new LinkedList<TransUnitDelegate>();

		for (final TransUnitDelegate transUnit : allTransUnitList) {
			final String state = transUnit.getTargetState();
			if (checkedStateItems.contains(state)) {
				transUnitList.add(transUnit);
			}
		}
		transUnitTableViewer.setInput(ElementDelegate.toElementList(transUnitList));

		if (transUnitList.size() == 0) {

			setEnabledTransUnit(false);

			noteTableViewer.setInput(Collections.EMPTY_LIST);
			idText.setText("");
			stateCombo.setText("");
			sourceTextViewer.getDocument().set("");
			targetTextViewer.getDocument().set("");

			clearSelectedAltTrans();
			updateAltTransButtons();

		} else if (transUnitTableViewer.getSelection().isEmpty()) {

			final StructuredSelection sel = new StructuredSelection(transUnitTableViewer.getElementAt(0));
			transUnitTableViewer.setSelection(sel, true);
			final IContentOutlinePage outline = (IContentOutlinePage) editorPart.getAdapter(IContentOutlinePage.class);
			outline.setSelection(sel);
		}
	}

	/**
	 * このビューアーの選択を取得します。
	 * このビューアーがアクティブになった時に呼び出されます。
	 * この選択はアウトライン・ビューなどに反映されます。
	 */
	@Override
	public ISelection getSelection() {

		refresh();

		CatUiPlugin.getDefault().setActiveEditorPart(editorPart);

		ISelection selection = altTransTableViewer.getSelection();
		if (selection.isEmpty()) {
			selection = transUnitTableViewer.getSelection();
		}
		return selection;
	}

	/**
	 * このビューアーの選択を設定します。
	 * ソース・タブやアウトライン・ビュー選択時に呼び出されます。
	 */
	@Override
	public void setSelection(ISelection selection, boolean reveal) {

		if (!selection.isEmpty() && selection instanceof IStructuredSelection) {

			final IStructuredSelection structuredSelection = (IStructuredSelection) selection;
			final Node firstElement = (Node) structuredSelection.getFirstElement();
			final Node node = firstElement;
			Node transUnitNode = null;
			Node altTransNode = null;
			Node noteNode = null;

			for (Node n = node; n != null; n = n.getParentNode()) {
				final String nodeName = n.getNodeName();
				if (noteNode == null && nodeName.equals("note")) {
					noteNode = n;
				} else if (altTransNode == null && nodeName.equals("alt-trans")) {
					altTransNode = n;
				} else if (transUnitNode == null && nodeName.equals("trans-unit")) {
					transUnitNode = n;
					break;
				}
			}
			try {
				internalSelectionChange = true;
				if (transUnitNode != null) {
					final ISelection sel = new StructuredSelection(transUnitNode);
					transUnitTableViewer.setSelection(sel, true);
				}
				if (altTransNode != null) {
					final ISelection sel = new StructuredSelection(altTransNode);
					altTransTableViewer.setSelection(sel, true);
				}
				if (noteNode != null) {
					final ISelection sel = new StructuredSelection(noteNode);
					noteTableViewer.setSelection(sel, true);
				}
			} finally {
				internalSelectionChange = false;
			}
		}
	}

	//-------------------------------------------------------------------------
	// org.eclipse.wst.xml.ui.internal.tabletree.XMLTableTreeViewer からのコピー

	private final ISelectionProvider selectionProvider = new SelectionProvider();

	private class SelectionProvider implements IPostSelectionProvider {

		public void addPostSelectionChangedListener(ISelectionChangedListener listener) {
			// 何もしない
			//XMLTableTreeViewer.this.addPostSelectionChangedListener(listener);
		}

		public void removePostSelectionChangedListener(ISelectionChangedListener listener) {
			// 何もしない
			//XMLTableTreeViewer.this.removePostSelectionChangedListener(listener);
		}

		public void addSelectionChangedListener(ISelectionChangedListener listener) {
			TransUnitViewer.this.addSelectionChangedListener(listener);
		}

		public ISelection getSelection() {
			return TransUnitViewer.this.getSelection();
		}

		public void removeSelectionChangedListener(ISelectionChangedListener listener) {
			TransUnitViewer.this.removeSelectionChangedListener(listener);
		}

		public void setSelection(ISelection selection) {
			boolean selectionSet = false;
			if (!selection.isEmpty() && selection instanceof IStructuredSelection) {
				final IStructuredSelection structuredSelection = (IStructuredSelection) selection;
				if (selection instanceof ITextSelection) {
					final ITextSelection textSelection = (ITextSelection) selection;

					if (structuredSelection.size() == 1) {
						if (structuredSelection.getFirstElement() instanceof IDOMNode) {
							final IDOMNode domNode = (IDOMNode) structuredSelection.getFirstElement();
							final IStructuredDocumentRegion startStructuredDocumentRegion = domNode
									.getStartStructuredDocumentRegion();
							if (startStructuredDocumentRegion != null) {
								ITextRegion matchingRegion = startStructuredDocumentRegion
										.getRegionAtCharacterOffset(textSelection.getOffset());
								int allowedIterations = 40;
								while (matchingRegion != null
										&& !matchingRegion.getType().equals(DOMRegionContext.XML_TAG_ATTRIBUTE_NAME)
										&& allowedIterations > 0) {
									matchingRegion = startStructuredDocumentRegion
											.getRegionAtCharacterOffset(startStructuredDocumentRegion
													.getStartOffset(matchingRegion) - 1);
									allowedIterations--;
								}
								if (matchingRegion != null
										&& matchingRegion.getType().equals(DOMRegionContext.XML_TAG_ATTRIBUTE_NAME)) {
									final String attrName = startStructuredDocumentRegion.getText(matchingRegion);
									final NamedNodeMap attributes = domNode.getAttributes();
									if (attributes != null && attrName.length() > 0) {
										final Node attr = attributes.getNamedItem(attrName);
										if (attr != null) {
											selectionSet = true;
											TransUnitViewer.this.setSelection(new StructuredSelection(attr));
										}
									}
								}
							}
						}
					}
				}
			}
			if (!selectionSet) {
				TransUnitViewer.this.setSelection(selection);
			}
		}
	}
}
